using Dates
using CodeTracking
using Git
using JLD2
using Logging
using Pkg.TOML

"""
    Experiment

This is a struct which encapsulates the idea of an experiment.

"""
struct Experiment{I}
    dir::AbstractString
    file::AbstractString
    module_name::Symbol
    func_name::Symbol
    args_iter::I
    hash::UInt64
    function Experiment(dir::AbstractString,
                        file::AbstractString,
                        module_name::Union{String, Symbol},
                        func_name::Union{String, Symbol},
                        args_iter)
        new{typeof(args_iter)}(dir, file, Symbol(module_name), Symbol(func_name), args_iter, hash(string(args_iter)))
    end
end

function Experiment(config::AbstractString)
    dict = TOML.parsefile(config)

    cdict = dict["config"]
    save_dir = cdict["save_dir"]
    exp_file = cdict["exp_file"]
    exp_module_name = cdict["exp_module_name"]
    exp_func_name = cdict["exp_func_name"]

    iter_type = cdict["arg_iter_type"]
    arg_list = get(cdict, "arg_list_order", nothing)

    @assert arg_list isa Nothing || all(sort(arg_list) .== sort(collect(keys(dict["sweep_args"]))))

    args = if iter_type == "iter"
        static_args_dict = get(dict, "static_args", Dict{String, Any}())
        static_args_dict["save_dir"] = joinpath(save_dir, "data")
        ArgIterator(dict["sweep_args"],
                    static_args_dict,
                    arg_list=arg_list)
    elseif iter_type == "looper"
        throw("Looper w/ Toml Config Not Implemented Yet")
    else
        throw("$(iter_type) not supported.")
    end

    experiment = Experiment(save_dir,
                            exp_file,
                            exp_module_name,
                            exp_func_name,
                            args)
    
end

function _safe_mkdir(exp_dir)
    try
        mkdir(exp_dir)
    catch ex
        @info "Somebody else created directory... Waiting"
        if isa(ex, SystemError) && ex.errnum == 17
            sleep(0.1) # Other Process Made folder. Waiting...
        else
            throw(ex)
        end
    end
end

"""
    create_experiment_dir(exp_dir::AbstractString; org_file=true, replace=false, tldr="")

This creates an experiment directory and sets up a notes file.

"""
function create_experiment_dir(exp_dir::AbstractString;
                               org_file=true,
                               replace=false,
                               tldr="")

    if isdir(exp_dir)
        if !replace
            @info "directory already created - told to not replace..."
            return
        else
            @info "directory already created - told to replace..."
            rm(exp_dir; force=true, recursive=true)
            _safe_mkdir(exp_dir)
        end
    else
        @info "creating experiment directory"
        _safe_mkdir(exp_dir)
    end

    if isdir(joinpath(exp_dir, "data"))
        if !replace
            @info "data directory already created - told to not replace..."
            return
        else
            @info "directory already created - told to replace..."
            # rm(exp_dir; force=true, recursive=true)
            _safe_mkdir(joinpath(exp_dir, "data"))
        end
    else
        @info "creating experiment directory"
        _safe_mkdir(joinpath(exp_dir, "data"))
    end

    if org_file
        open(joinpath(exp_dir, "notes.org"), "w") do f
            write(f, "#+title: Experimental Notes for $(exp_dir)\n\n")
            write(f, "TL;DR: $(tldr)\n\n")
        end
    end

    return
end

function create_experiment_dir(exp::Experiment;
                               kwargs...)
    create_experiment_dir(exp.dir; kwargs...)
    return
end


function add_experiment(exp_dir::AbstractString,
                        experiment_file::AbstractString,
                        exp_module_name::AbstractString,
                        exp_func_name::AbstractString,
                        args_iter::AbstractArgIter,
                        hash::UInt64;
                        settings_dir="", add_all_tasks=false)

    if "SLURM_ARRAY_TASK_ID" in keys(ENV)
        if parse(Int64, ENV["SLURM_ARRAY_TASK_ID"]) != 1 && !add_all_tasks
            job_id = parse(Int64, ENV["SLURM_ARRAY_TASK_ID"])
            @info "Told to not add all experiments... job_id : $(job_id) $(job_id == 1)"
            return
        end
    end

    @info "Adding Experiment to $(exp_dir)"

    settings_dir = joinpath(exp_dir, settings_dir)
    _safe_mkdir(settings_dir)

    settings_file = joinpath(settings_dir, "settings_0x"*string(hash, base=16)*".jld2")

    date_str = Dates.format(now(), dateformat"<yyyy-mm-dd e HH:MM:SS>")
    tab = "\t"

    make_args_str = "nothing"
    if typeof(args_iter) == ArgIterator && args_iter.make_args != nothing
        m = CodeTracking.@which args_iter.make_args(Dict{String, String}())
        make_args_str, line1 = definition(String, m)
    end

    open(joinpath(exp_dir, "notes.org"), "a") do f
        exp_str = "* " * date_str * "\n\n" *
            tab*"Git-head: $(Git.head())\n" *
            tab*"Git-branch: $(Git.branch())\n" *
            tab*"experiment file: $(experiment_file)\n" *
            tab*"experiment module: $(string(exp_module_name))\n" *
            tab*"experiment function: $(string(exp_func_name))\n\n" *
            tab*"settings file: $(settings_dir)\n\n" *
            tab*"#+BEGIN_SRC julia\n" *
            (args_iter isa ArgIterator ? tab*"dict = $(args_iter.dict)\n" : tab*"runs_iter=$(args_iter.runs_iter)\n") *
            (args_iter isa ArgIterator ? tab*"arg_list = $(args_iter.arg_list)\n" : tab*"arg_list = $(args_iter.dict_list)\n") *
            tab*"static_arg = $(args_iter.static_args)\n\n" *
            tab*"#Make Arguments\n" *
            tab*make_args_str*"\n" *
            tab*"#+END_SRC\n\n"
        write(f, exp_str)
    end

    jldopen(settings_file, "w") do file
        file["args_iter"]=args_iter
        file["make_args_str"]=make_args_str
    end

end

function add_experiment(exp::Experiment;
                        kwargs...)
    add_experiment(exp.dir,
                   exp.file,
                   String(exp.module_name),
                   String(exp.func_name),
                   exp.args_iter,
                   exp.hash;
                   kwargs...)
end

function post_experiment(exp_dir::AbstractString, canceled_jobs::Array{Int64, 1})

    if "SLURM_ARRAY_TASK_ID" in keys(ENV)
        @info "Post_experiment doesn't work with slurm job arrays."
        return
    end

    tab = "\t"
    date_str = Dates.format(now(), dateformat"<yyyy-mm-dd e HH:MM:SS>")
    open(joinpath(exp_dir, "notes.org"), "a") do f

        post_exp_str = tab*"Post Experiment: \n" *
            tab*"Canceled Jobs: $(canceled_jobs)\n" *
            tab*"Ended: $(date_str)\n"
        write(f, post_exp_str)
    end
end

function post_experiment(exp_dir::AbstractString, finished_job::Bool)
    @info "Post_experiment not supported with task jobs."
end

function post_experiment(exp::Experiment, job_ret)
    post_experiment(exp.dir, job_ret)
end

function exception_file(exc_file::AbstractString, job_id, exception, trace)

    if isfile(exc_file)
        @warn "$(exc_file) already exists. Overwriting..."
    end

    open(exc_file, "w") do f
        exception_string =
            "Exception for job_id: $(job_id)\n\n" * string(exception) * "\n\n"

        write(f, exception_string)
        Base.show_backtrace(f, trace)
    end

    return
end

